#include "deftypes.h"
#include "raine.h"
#include "priorities.h"

extern int disp_x_16;

void ldraw16x16_Mapped(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;
   for(yy=0; yy<16; yy++){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(GameBitmap->line[y+yy]))+ dx;
      for(xx=0; xx<16; xx++){
         *line++ = ((UINT8 *)cmap)[ *SPR++ ];
      }
   }
}

void ldraw16x16_Mask(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;
   UINT32 mask = cmap | (cmap << 8);
   mask |= (mask << 16);

   for(yy=0; yy<16; yy++){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       continue;
     }
     line = pbitmap->line[y+yy]+ dx;
      for(xx=0; xx<16; xx+=4){
         WriteLong(&line[xx],mask);
      }
   }
}

// The column draw is designed to replace the line scroll when the screen
// is rotated by 270. If a game uses 90 or 180 we will use 2 other famillies
// of functions.
// This cdraw16x16 is very unefficient compared to the usual Draw16x16
// functions because :
// - the line scroll indexes follow y (on screen)
// - the sprite follows x.
// This whole thing should probably be entierly rewritten, and preferably using
// a macro this time to have only 1 version for 8,16 and 32 bpp.

void cdraw16x16_Mapped(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = GameBitmap->w;

#if 1
   for(xx=0; xx<=15; xx++){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[xx];
     line = ((UINT8 *)(GameBitmap->line[dy]))+x+xx ;
     for(yy=0; yy<16; yy++){
       line[yy*w] = cmap[ spr[yy*16] ];
     }
   }
#else
   for (yy=0; yy<16; yy++) {
     line = GameBitmap->line[yy]+ x;
     for (xx=0; xx<16; xx++) {
       dy = y-dx[xx];
       if (dy < 16 || dy > disp_x_16) continue;
       line[xx+dy*w] = cmap[ *SPR++ ];
     }
   }
#endif
}

void cdraw16x16_Mask(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line;
   INT16 xx,yy,dy;
   UINT32 w;
   w = pbitmap->w;

#if 1
   for(xx=0; xx<=15; xx++){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       // fprintf(stderr,"%d > %d (%d)\n",dy,disp_x_16,disp_x_16);
       continue;
     }
     line = ((UINT8 *)(pbitmap->line[dy]))+x + xx ;
     for(yy=0; yy<16; yy++){
       line[yy*w] = cmap;
     }
   }
#else
   for (yy=0; yy<16; yy++) {
     line = pbitmap->line[yy]+ x;
     for (xx=0; xx<16; xx++) {
       dy = y-dx[xx];
       if (dy < 16 || dy > disp_x_16) continue;
       line[xx+dy*w] = cmap;
     }
   }
#endif

}

void ldraw16x16_Mapped_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=15; yy>=0; yy--){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(GameBitmap->line[y+yy]))+ dx;
     for(xx=0; xx<16; xx++){
       *line++ = ((UINT8 *)cmap)[ *SPR++ ];
     }
   }
}

void cdraw16x16_Mapped_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = GameBitmap->w;

   for(xx=0; xx<=15; xx++){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[xx];
     line = ((UINT8 *)(GameBitmap->line[dy]))+x+xx ;
     for(yy=15; yy>=0; yy--){
       line[yy*w] = ((UINT8 *)cmap)[ spr[16*(15-yy)] ];
     }
   }
}

void ldraw16x16_Mapped_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=0; yy<16; yy++){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(GameBitmap->line[y+yy])) + dx+15;
     for(xx=0; xx<16; xx++){
       *line-- = ((UINT8 *)cmap)[ *SPR++ ];
     }
   }
}

void cdraw16x16_Mapped_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = GameBitmap->w;
   for(xx=15; xx>=0; xx--){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[15-xx];
     line = ((UINT8 *)(GameBitmap->line[dy]))+x+xx ;
     for(yy=0; yy<16; yy++){
       line[yy*w] = ((UINT8 *)cmap)[ spr[16*yy] ];
     }
   }
}

void ldraw16x16_Mapped_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=15; yy>=0; yy--){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(GameBitmap->line[y+yy])) + dx + 15;
     for(xx=0; xx<16; xx++){
       *line-- = ((UINT8 *)cmap)[ *SPR++ ];
     }
   }
}

void cdraw16x16_Mapped_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = GameBitmap->w;
   for(xx=15; xx>=0; xx--){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[15-xx];
     line = ((UINT8 *)(GameBitmap->line[dy]))+x+xx ;
     for(yy=15; yy>=0; yy--){
       line[yy*w] = ((UINT8 *)cmap)[ spr[16*(15-yy)] ];
     }
   }
}

// 16x16 mapped transparent sprites

void ldraw16x16_Trans_Mapped(UINT8 *SPR, int x, int y, UINT8 *cmap,INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=0; yy<16; yy++){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(GameBitmap->line[y+yy])) + dx;
     for(xx=0; xx<16; xx++, SPR++){
       if(*SPR)
	 line[xx] = ((UINT8 *)cmap)[ *SPR ];
     }
   }
}

void ldraw16x16_Mask_Trans(UINT8 *SPR, int x, int y, UINT8 cmap,INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=0; yy<16; yy++){
     dx = x+dy[yy];
     if (dx < 16 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = pbitmap->line[y+yy] + dx;
     for(xx=0; xx<16; xx++, SPR++){
       if(*SPR)
	 line[xx] = cmap;
     }
   }
}

void cdraw16x16_Trans_Mapped(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = GameBitmap->w;
   for(xx=0; xx<=15; xx++){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[xx];
     line = ((UINT8 *)(GameBitmap->line[dy]))+x+xx ;
     for(yy=0; yy<16; yy++){
       UINT8 myspr = spr[yy*16];
       if(myspr) {
	 line[yy*w] = ((UINT8 *)cmap)[ myspr ];
       }
     }
   }
}

void cdraw16x16_Mask_Trans(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = pbitmap->w;
   for(xx=0; xx<=15; xx++){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[xx];
     line = ((UINT8 *)(pbitmap->line[dy]))+x+xx ;
     for(yy=0; yy<16; yy++){
       if(spr[yy*16]) {
	 line[yy*w] = cmap;
       }
     }
   }
}

void ldraw16x16_Trans_Mapped_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=15; yy>=0; yy--){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(GameBitmap->line[y+yy])) + dx;
     for(xx=0; xx<16; xx++, SPR++){
       if(*SPR)
	 line[xx] = ((UINT8 *)cmap)[ *SPR ];
     }
   }
}

void ldraw16x16_Mask_Trans_FlipX(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=15; yy>=0; yy--){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(pbitmap->line[y+yy])) + dx;
     for(xx=0; xx<16; xx++, SPR++){
       if(*SPR)
	 line[xx] = cmap;
     }
   }
}

void cdraw16x16_Trans_Mapped_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = GameBitmap->w;

   for(xx=0; xx<=15; xx++){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[xx];
     line = ((UINT8 *)(GameBitmap->line[dy]))+x+xx ;
     for(yy=15; yy>=0; yy--){
       UINT8 myspr = spr[16*(15-yy)];
       if (myspr) {
	 line[yy*w] = cmap[ myspr ];
       }
     }
   }
}

void cdraw16x16_Mask_Trans_FlipX(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = pbitmap->w;

   for(xx=0; xx<=15; xx++){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[xx];
     line = ((UINT8 *)(pbitmap->line[dy]))+x+xx ;
     for(yy=15; yy>=0; yy--){
       if (spr[16*(15-yy)]) {
	 line[yy*w] = cmap;
       }
     }
   }
}

void ldraw16x16_Trans_Mapped_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=0; yy<16; yy++){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(GameBitmap->line[y+yy])) + dx;
     for(xx=15; xx>=0; xx--, SPR++){
       if(*SPR)
	 line[xx] = ((UINT8 *)cmap)[ *SPR ];
     }
   }
}

void ldraw16x16_Mask_Trans_FlipY(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=0; yy<16; yy++){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(pbitmap->line[y+yy])) + dx;
     for(xx=15; xx>=0; xx--, SPR++){
       if(*SPR)
	 line[xx] = cmap;
     }
   }
}

void cdraw16x16_Trans_Mapped_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = GameBitmap->w;

   for(xx=15; xx>=0; xx--){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[15-xx];
     line = ((UINT8 *)(GameBitmap->line[dy+yy]))+x+xx ;
     for(yy=0; yy<16; yy++){
       UINT8 myspr = spr[16*(yy)];
       if (myspr) {
	 line[yy*w] = cmap[ myspr ];
       }
     }
   }
}

void cdraw16x16_Mask_Trans_FlipY(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = pbitmap->w;

   for(xx=15; xx>=0; xx--){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[15-xx];
     line = ((UINT8 *)(pbitmap->line[dy]))+x+xx ;
     for(yy=0; yy<16; yy++){
       if (spr[16*(yy)]) {
	 line[yy*w] = cmap;
       }
     }
   }
}

void ldraw16x16_Trans_Mapped_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=15; yy>=0; yy--){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(GameBitmap->line[y+yy])) + dx;
     for(xx=15; xx>=0; xx--, SPR++){
       if(*SPR)
	 line[xx] = ((UINT8 *)cmap)[ *SPR ];
     }
   }
}

void ldraw16x16_Mask_Trans_FlipXY(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy,dx;

   for(yy=15; yy>=0; yy--){
     dx = x+dy[yy];
     if (dx < 0 || dx > disp_x_16) {
       SPR += 16;
       continue;
     }
     line = ((UINT8 *)(pbitmap->line[y+yy])) + dx;
     for(xx=15; xx>=0; xx--, SPR++){
       if(*SPR)
	 line[xx] = cmap;
     }
   }
}

void cdraw16x16_Trans_Mapped_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;

   for(xx=15; xx>=0; xx--){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[15-xx];
     line = ((UINT8 *)(GameBitmap->line[dy]))+x+xx ;
     for(yy=15; yy>=0; yy--){
       UINT8 myspr = spr[16*(15-yy)];
       if (myspr) {
	 line[yy*16] = cmap[ myspr ];
       }
     }
   }
}

void cdraw16x16_Mask_Trans_FlipXY(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dx)
{
  /* Column draw, or line scroll rotated by 270 */
   UINT8 *line,*spr;
   INT16 xx,yy,dy;
   UINT32 w = pbitmap->w;

   for(xx=15; xx>=0; xx--){
     dy = y-dx[xx];
     if (dy < 16 || dy > disp_x_16) {
       continue;
     }
     spr = &SPR[15-xx];
     line = ((UINT8 *)(pbitmap->line[dy]))+x+xx ;
     for(yy=15; yy>=0; yy--){
       if (spr[16*(15-yy)]) {
	 line[yy*w] = cmap;
       }
     }
   }
}

// 16x16 sprites (I wonder if these will ever be used).
// Since I don't feel courageous, I am not going to write the bound
// checking code for real sprites...

// This is really a piece of history. The only game I know which uses
// direct mapping is bubble bobble !
#if 0
// commented until we find they are usefull (unlikely)
void ldraw16x16(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT8 *)(GameBitmap->line[y+yy])) + x + dy[yy];
      for(xx=0; xx<16; xx++){
         *line++ = *SPR++ | cmap;
      }
   }
}

void ldraw16x16_FlipX(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT8 *)(GameBitmap->line[y+yy])) + x + dy[yy];
      for(xx=0; xx<16; xx++){
         *line++ = *SPR++ | cmap;
      }
   }
}

void ldraw16x16_FlipY(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT8 *)(GameBitmap->line[y+yy])) + x + 7 + dy[yy];
      for(xx=0; xx<16; xx++){
         *line-- = *SPR++ | cmap;
      }
   }
}

void ldraw16x16_FlipXY(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT8 *)(GameBitmap->line[y+yy])) + x + 7 + dy[yy];
      for(xx=0; xx<16; xx++){
         *line-- = *SPR++ | cmap;
      }
   }
}

// 16x16 transparent sprites

void ldraw16x16_Trans(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT8 *)(GameBitmap->line[y+yy])) + x + dy[yy];
      for(xx=0; xx<16; xx++, SPR++){
         if(*SPR)
            line[xx] = *SPR | cmap;
      }
   }
}

void ldraw16x16_Trans_FlipX(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT8 *)(GameBitmap->line[y+yy])) + x + dy[yy];
      for(xx=0; xx<16; xx++, SPR++){
         if(*SPR)
            line[xx] = *SPR | cmap;
      }
   }
}

void ldraw16x16_Trans_FlipY(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT8 *)(GameBitmap->line[y+yy])) + x + dy[yy];
      for(xx=15; xx>=0; xx--, SPR++){
         if(*SPR)
            line[xx] = *SPR | cmap;
      }
   }
}

void ldraw16x16_Trans_FlipXY(UINT8 *SPR, int x, int y, UINT8 cmap, INT16 *dy)
{
   UINT8 *line;
   INT16 xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT8 *)(GameBitmap->line[y+yy])) + x + dy[yy];
      for(xx=15; xx>=0; xx--, SPR++){
         if(*SPR)
            line[xx] = *SPR | cmap;
      }
   }
}
#endif
