/******************************************************************************/
/*                                                                            */
/*                         RAINE 16x16 TILE DRAWING                           */
/*                                                                            */
/******************************************************************************/

#include "deftypes.h"
#include "raine.h"
#include "priorities.h"

/*

Mapped
Transparent Mapped
Direct-Mapped
Transparent Direct-Mapped

*/

// taken from 16x16_8.c. Difference : the priority bitmap
// px, py are passed to avoid the conversion of coordinates because of screen rotation
/* See the comments about pdraw and pbitmap usage in priorities.c */
// 16x16 mapped sprites

void pdraw16x16_Mapped(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT8 *line,*pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = GameBitmap->line[y+yy] + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
	if (pline[xx] < pri)
	  line[xx] = cmap[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Mapped_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT8 *line,*pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = GameBitmap->line[y+yy] + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
	if (pline[xx] < pri)
	  line[xx] = cmap[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Mapped_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT8 *line,*pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = GameBitmap->line[y+yy] + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
	if (pline[xx] < pri)
	  line[xx] = cmap[ *SPR ];
      }
   }
}

void pdraw16x16_Mapped_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT8 *line,*pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = GameBitmap->line[y+yy] + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
	if (pline[xx] < pri)
	  line[xx] = cmap[ *SPR ];
      }
   }
}

// 16x16 mapped transparent sprites

void pdraw16x16_Trans_Mapped(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT8 *line,*pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = GameBitmap->line[y+yy] + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
         if(SPR[xx] && pline[xx] < pri)
            line[xx] = cmap[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Trans_Mapped_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT8 *line,*pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = GameBitmap->line[y+yy] + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
         if(SPR[xx] && pline[xx] < pri)
            line[xx] = cmap[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Trans_Mapped_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT8 *line,*pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = GameBitmap->line[y+yy] + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
         if(*SPR && pline[xx] < pri)
            line[xx] = cmap[ *SPR ];
      }
   }
}

void pdraw16x16_Trans_Mapped_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT8 *line,*pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = GameBitmap->line[y+yy] + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
         if(*SPR && pline[xx] < pri)
            line[xx] = cmap[ *SPR ];
      }
   }
}

// 16bpp

void pdraw16x16_Mapped_16(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT16 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT16 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
	if (pline[xx] < pri)
	  line[xx] = ((UINT16 *)cmap)[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Mapped_16_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT16 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT16 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
	if (pline[xx] < pri)
	  line[xx] = ((UINT16 *)cmap)[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Mapped_16_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT16 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT16 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
	if (pline[xx] < pri)
	  line[xx] = ((UINT16 *)cmap)[ *SPR ];
      }
   }
}

void pdraw16x16_Mapped_16_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT16 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT16 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
	if (pline[xx] < pri)
	  line[xx] = ((UINT16 *)cmap)[ *SPR ];
      }
   }
}

// 16x16 mapped transparent sprites

void pdraw16x16_Trans_Mapped_16(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT16 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT16 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
         if(SPR[xx] && pline[xx] < pri)
            line[xx] = ((UINT16 *)cmap)[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Trans_Mapped_16_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT16 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT16 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
         if(SPR[xx] && pline[xx] < pri)
            line[xx] = ((UINT16 *)cmap)[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Trans_Mapped_16_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT16 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT16 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
         if(*SPR && pline[xx] < pri)
            line[xx] = ((UINT16 *)cmap)[ *SPR ];
      }
   }
}

void pdraw16x16_Trans_Mapped_16_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT16 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT16 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
         if(*SPR && pline[xx] < pri)
            line[xx] = ((UINT16 *)cmap)[ *SPR ];
      }
   }
}

// 32 bpp

void pdraw16x16_Mapped_32(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT32 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT32 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
	if (pline[xx] < pri)
	  line[xx] = ((UINT32 *)cmap)[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Mapped_32_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT32 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT32 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
	if (pline[xx] < pri)
	  line[xx] = ((UINT32 *)cmap)[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Mapped_32_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT32 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT32 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
	if (pline[xx] < pri)
	  line[xx] = ((UINT32 *)cmap)[ *SPR ];
      }
   }
}

void pdraw16x16_Mapped_32_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT32 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT32 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
	if (pline[xx] < pri)
	  line[xx] = ((UINT32 *)cmap)[ *SPR ];
      }
   }
}

// 16x16 mapped transparent sprites

void pdraw16x16_Trans_Mapped_32(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT32 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT32 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
         if(SPR[xx] && pline[xx] < pri)
            line[xx] = ((UINT32 *)cmap)[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Trans_Mapped_32_FlipX(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT32 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT32 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=0; xx<16; xx++){
         if(SPR[xx] && pline[xx] < pri)
            line[xx] = ((UINT32 *)cmap)[ SPR[xx] ];
      }
      SPR += 16;
   }
}

void pdraw16x16_Trans_Mapped_32_FlipY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT32 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=0; yy<16; yy++){
      line = ((UINT32 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
         if(*SPR && pline[xx] < pri)
            line[xx] = ((UINT32 *)cmap)[ *SPR ];
      }
   }
}

void pdraw16x16_Trans_Mapped_32_FlipXY(UINT8 *SPR, int x, int y, UINT8 *cmap,  int pri)
{
   UINT32 *line;
   UINT8 *pline;
   int xx,yy;

   for(yy=15; yy>=0; yy--){
      line = ((UINT32 *)(GameBitmap->line[y+yy])) + x;
      pline = pbitmap->line[y+yy] + x;
      for(xx=15; xx>=0; xx--,SPR++){
         if(*SPR && pline[xx] < pri)
            line[xx] = ((UINT32 *)cmap)[ *SPR ];
      }
   }
}

